#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import QUB
from Buttons import *
from Dialogs import *
from qubtree import Node
import math
from itertools import *

def children(tree, name=None):
    if name == None:
        child = tree.child
    else:
        child = tree.find(name)
    while not child.isNull:
        yield child
        if name == None:
            child = child.sibling
        else:
            child = child.nextSameName()


def AppendModelTrial(result, modelObj, name):
    milResult = QUB.ResultsView.currResults.node
    trial = result.append('ModelTrial')
    trial.appendClone(milResult['LL']).name = ' LL'
    if not milResult.find('Gradient').isNull:
        trial.appendClone(milResult['Gradient'])
    else:
        trial.appendClone(milResult['Free pars grads']).name = 'Gradient' # MIP style
    mdl = trial.append(modelObj.getAsNode())
    mdl.data = name
    ds = milResult.find('DataSet')
    while not ds.isNull:
        trial.appendClone(ds)
        ds = ds.nextSameName()


def CountPrtbs(model):
    fixK0 = [tuple(c.data[:]) for c in children(model['Constraints'], 'FixRate')]
    fixK1 = [tuple(c.data[:]) for c in children(model['Constraints'], 'FixExp')]
    count = 0
    for rate in children(model['Rates'], 'Rate'):    
        a, b = rate['States'].data[:2]
        states = [(a,b), (b,a)]
        for dir in range(2):
            if states[dir] in fixK0:
                continue
            count += 2
        for dir in range(2):
            origK1 = rate['k1'].data[dir]
            if (not rate['Q'].data[dir]) or (states[dir] in fixK1) or (not origK1):
                continue
            count += 2
    return count    

def UpdatePrtbCount():
    model = QUB.ModelView.currModel
    QUB.Profiles['Prtb'].active['count'].data = model and CountPrtbs(model.getAsNode()) or 0

def PerturbEachRate(props):
#    try:
        pct = props['pct']
        logpctmult = math.log(1.0 + 0.01 * pct)
        
        modelObj = QUB.ModelView.currModel
        if not modelObj:
            QUB.Report('no model')
            return
        model = modelObj.getAsNode()

        result = Node()
        result['ResultType'].data = 'Prtb'

        newMilSettings = Node()
        newMilSettings['Mode'].data = 'evaluate'
        newMilSettings['use segments'].data = 'together'

        Run( 'Modeling', props['Algorithm'], newMilSettings )
        AppendModelTrial(result, modelObj, 'Initial')

        newMilSettings['Mode'].data = str(props['Mode'])

        fixK0 = [tuple(c.data[:]) for c in children(model['Constraints'], 'FixRate')]
        fixK1 = [tuple(c.data[:]) for c in children(model['Constraints'], 'FixExp')]

        for rate in children(model['Rates'], 'Rate'):    
            a, b = rate['States'].data[:2]
            states = [(a,b), (b,a)]
            for dir in range(2):
                if states[dir] in fixK0:
                    continue

                origK0 = math.log( rate['k0'].data[dir] )
                for offsetdir in range(-1, 2, 2):
                    rate['k0'].data[dir] = math.exp(origK0 + offsetdir * logpctmult) # exp(origK0 * (1.0 + .01 * offset))
                    modelObj.setAsNode( model )
                    Run( 'Modeling', props['Algorithm'], newMilSettings )
                    AppendModelTrial( result, modelObj, '%d-%d  %d%%' % (states[dir][0]+1, states[dir][1]+1, offsetdir * pct) )
                rate['k0'].data[dir] = math.exp(origK0)
            for dir in range(2):
                origK1 = rate['k1'].data[dir]
                if (not rate['Q'].data[dir]) or (states[dir] in fixK1) or (not origK1):
                    continue
                xsgn = (origK1 > 0) and 1.0 or -1.0
                logK1 = math.log(xsgn * origK1)
                for offsetdir in range(-1, 2, 2):
                    rate['k1'].data[dir] = xsgn * math.exp(logK1 + offsetdir * logpctmult) # * (1.0 + offsetdir * pct/100.0) # origK1 + offsetdir * logpctmult # exp(origK0 * (1.0 + .01 * offset))
                    modelObj.setAsNode( model )
                    Run( 'Modeling', props['Algorithm'], newMilSettings )
                    AppendModelTrial( result, modelObj, 'k1 %d-%d  %d%%' % (states[dir][0]+1, states[dir][1]+1, xsgn * offsetdir * pct) )
                rate['k1'].data[dir] = origK1

        if props['Algorithm'] == 'Mac':
            macProps = QUB.Modeling['Mac'].Properties()
            if macProps['Est']['ChannelCount'].data[0]:
                cc = model['ChannelCount'].data[0]
                mag = max(1, int(pct * cc / 100.0))
                for sgn in range(-1, 2, 2):
                    model['ChannelCount'].data[0] = max(1, cc + sgn*mag)
                    modelObj.setAsNode(model)
                    Run('Modeling', 'Mac', newMilSettings)
                    AppendModelTrial(result, modelObj, 'ChannelCount %d%%' % (sgn*pct))
                model['ChannelCount'].data[0] = cc
            if macProps['Est']['AmpOpen'].data[0]:
                x = model['Amps'].data[1] or 1e-8
                xsgn = (x > 0) and 1.0 or -1.0
                x = math.log( x * xsgn )
                for sgn in range(-1, 2, 2):
                    model['Amps'].data[1] = xsgn * math.exp(x + sgn*logpctmult)
                    modelObj.setAsNode(model)
                    Run('Modeling', 'Mac', newMilSettings)
                    AppendModelTrial(result, modelObj, 'AmpOpen %d%%' % (xsgn*sgn*pct))
                model['Amps'].data[1] = xsgn * math.exp(x)
            if macProps['Est']['VarOpen'].data[0]:
                x = model['Stds'].data[1] or 1e-8
                xsgn = (x > 0) and 1.0 or -1.0
                x = math.log( x * xsgn )
                for sgn in range(-1, 2, 2):
                    model['Stds'].data[1] = xsgn * math.exp(x + sgn*logpctmult)
                    modelObj.setAsNode(model)
                    Run('Modeling', 'Mac', newMilSettings)
                    AppendModelTrial(result, modelObj, 'VarOpen %d%%' % (xsgn*sgn*pct))
                model['Stds'].data[1] = xsgn * math.exp(x)

        modelObj.setAsNode( model )
#    except:
#        pass
    
        QUB.ResultsView.addResults( result )
        


AddTaskButton( 'Perturb Rates', # button text
               PerturbEachRate,   # function
               'Prtb',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Perturb each rate then run M__ to verify a model",   # tool-tip
               "Perturb each rate then run M__ to verify a model",   # description in "Edit Buttons"
               "Perturb Rates properties:", # caption for the settings dialog
               [DialogText("pct", "Percent perturbation:", 10, acceptIntGreaterThan(0)),
                DialogRadioGroup("Algorithm", "Algorithm:", ['MIL', 'MSL', 'MIP', 'Mac'], 'MIL'),
                DialogRadioGroup("Mode", "Mode:", ['optimize', 'evaluate'], 'evaluate'),
                DialogText("count", "Number of perturbations:", 1, acceptIntGreaterThan(-1))],
               'Modeling', onDlogShow=UpdatePrtbCount,
               DataChannel=True, DataList=True)
